/*
 * Copyright © 2006 Carl Worth
 *
 * This program is free software; you can redistribute it and\/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA."
 */

#include "word-game.h"

#include <sys/time.h>
#include <time.h>
#include <math.h>

#include <readline/readline.h>
#include <readline/history.h>

/* Remember that dict reserves the 0th bit for IS_WORD */
#define GAME_WORD_SEEN (1<<1)

static bool_t
seen_predicate (dict_entry_t entry)
{
    return entry & GAME_WORD_SEEN;
}

static bool_t
unseen_predicate (dict_entry_t entry)
{
    return ! seen_predicate (entry);
}

static void
_count_possible (dict_cursor_t	cursor,
		 int		depth,
		 int		possible[17])
{
    char c;

    if (cursor == DICT_CURSOR_NIL)
	return;

    if (DICT_ENTRY_IS_WORD (dict_cursor_resolve (cursor)))
	possible[depth]++;

    for (c = 'a'; c <= 'z'; c++)
	_count_possible (dict_cursor_next (cursor, c), depth + 1, possible);
}

static void
_print_stats (int found[WORD_GAME_MAX_WORD_LENGTH+1],
	      int possible[WORD_GAME_MAX_WORD_LENGTH+1],
	      int found_total,
	      int possible_total)
{
    int i;

    for (i = 2; i <= 17; i++)
	if (possible[i])
	    printf ("%d:(%d/%d) ", i, found[i], possible[i]);

    printf ("total:(%d/%d = %.2f%%)\n",
	    found_total, possible_total,
	    100 * (double) found_total / possible_total);
}


void
word_game_play (const char	*puzzle,
		dict_t		*dict,
		dict_t		*answers,
		int		 time_limit_seconds)
{
    int i;
    struct timeval tv, tv_stop;
    int remaining, minutes, seconds;
    char prompt[7], *response;
    bool_t just_saw_puzzle;
    int found[WORD_GAME_MAX_WORD_LENGTH+1];
    int possible[WORD_GAME_MAX_WORD_LENGTH+1];
    int found_total, possible_total;

    for (i=0; i < WORD_GAME_MAX_WORD_LENGTH+1; i++) {
	found[i] = 0;
	possible[i] = 0;
    }
    _count_possible (dict_root (answers), 0, possible);
    found_total = 0;
    possible_total = 0;
    for (i=0; i < 17; i++)
	possible_total += possible[i];

    printf ("%s\n", puzzle);
    just_saw_puzzle = TRUE;

    remaining = time_limit_seconds;
    if (time_limit_seconds) {
	gettimeofday (&tv, NULL);
	tv_stop = tv;
	tv_stop.tv_sec += time_limit_seconds;
    } else {
	sprintf (prompt, "> ");
	remaining = 1;
    }
    do {
	if (time_limit_seconds) {
	    minutes = remaining / 60;
	    seconds = remaining % 60;
	    sprintf (prompt, "%02d:%02d ", minutes, seconds);
	}
	response = readline (prompt);
	if (response == NULL)
	    break;
	if (strlen (response) == 0) {
	    if (! just_saw_puzzle) {
		printf ("%s\n", puzzle);
		just_saw_puzzle = TRUE;
	    } else {
		_print_stats (found, possible, found_total, possible_total);
	    }
	} else {
	    dict_entry_t *entry;
	    just_saw_puzzle = FALSE;
	    add_history (response);
	    if (response[strlen (response) - 1] == '\n')
		response[strlen (response) - 1] = '\0';
	    entry = dict_lookup (answers, response);
	    if (DICT_ENTRY_IS_WORD (entry)) {
		if (*entry & GAME_WORD_SEEN) {
		    printf ("(repeat)\n");
		} else {
		    *entry |= GAME_WORD_SEEN;
		    found [strlen (response)]++;
		    found_total++;
		}
	    } else {
		entry = dict_lookup (dict, response);
		if (DICT_ENTRY_IS_WORD (entry))
		    printf ("(a good word, but it's not in the puzzle)\n");
		else
		    printf ("*** %s is not a word\n", response);
	    }
	}
	free (response);
	if (time_limit_seconds) {
	    gettimeofday (&tv, NULL);
	    remaining = floor (0.5 + (tv_stop.tv_sec - tv.tv_sec) + (tv_stop.tv_usec - tv.tv_usec) / 1000000.0);
	    minutes = remaining / 60;
	}
    } while (remaining > 0);

    printf ("\n%s\n", puzzle);

    printf ("\n");
    _print_stats (found, possible, found_total, possible_total);
    printf ("\n");

    if (found_total) {
	printf ("Words you found:\n");
	dict_print_by_length_if (answers, seen_predicate);
	printf ("\n");
    }

    if (found_total < possible_total) {
	printf ("Words you missed:\n");
	dict_print_by_length_if (answers, unseen_predicate);
	printf ("\n");
    }

    printf ("You found %d of %d words (%.2f%%)\n",
	    found_total, possible_total,
	    100 * (double) found_total / possible_total);
}
